/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <wchar.h>

#include <iomanip>

#include <stdlib.h>
#include <getopt.h>

#include "libtkcommon/config.h"
#include "libtkcommon/tkintl.h"
#include "libtkanalysis/analyser.h"
#include "libtkdb/database.h"

using namespace TkResolver;
using namespace Analysis;

void
print_help ()
{
  std::cout  << PACKAGE_NAME << " version " << PACKAGE_VERSION << std::endl;
  std::wcout << L"Copyright (C) 2005 Robert Ham (rah@bash.sh)" << std::endl;
  std::wcout << std::endl;
  std::wcout << L"This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it" << std::endl;
  std::wcout << L"under the terms of the GNU General Public License, version 2 or later.  See" << std::endl;
  std::wcout << L"the COPYING file that came with this software for details." << std::endl;
  std::wcout << std::endl;
  std::wcout << L" -h, --help                  " << _("Display this help info") << std::endl;
  std::wcout << L" -d, --db-host <host>        " << _("Connect to database on <host>") << std::endl;
  std::wcout << L" -g, --group <name>          " << _("Experiment group name") << std::endl;
  std::wcout << L" -p, --pseudo                " << _("Results with pseudo random data") << std::endl;
  std::wcout << L" -c, --control               " << _("Control results") << std::endl;
  std::wcout << std::endl;
}

int
main (int argc, char ** argv)
{
  TkString group_name("Test group");
  TkString db_host("localhost");
  TkString algorithm;
  bool algorithm_details(false);
  bool pseudo(false);
  bool control(false);
  wchar_t * wstr;

  int opt;
  const char * options = "hg:pcd:";
  struct option long_options[] = {
    { "help",        no_argument,       NULL, 'h' },
    { "group",       required_argument, NULL, 'g' },
    { "pseudo",      no_argument,       NULL, 'p' },
    { "control",     no_argument,       NULL, 'c' },
    { "db-host",     required_argument, NULL, 'd' },
    { 0, 0, 0, 0 }
  };
  
  while ((opt = getopt_long (argc, argv, options, long_options, NULL)) != -1) {
    switch (opt) {
      
    case 'g':
      group_name = optarg;
      break;

    case 'p':
      pseudo = true;
      break;

    case 'c':
      control = true;
      break;

    case 'd':
      db_host = optarg;
      break;

    case 'h':
      print_help ();
      exit (0);
      break;
    case ':':
    case '?':
      print_help ();
      exit (1);
      break;
    }
  }


  try
    {
      Db::Database database(db_host);
	  
      std::wcout << _(L"Analysing results for experiment group")
		 << " '"
		 << group_name
		 << "'..."
		 << std::flush;
	  
      Analyser analyser(database, group_name, !pseudo, control);
	  
      std::wcout << _(L"done")
		 << std::endl
		 << std::endl;



      Count n(analyser.get_n());

      std::wcout << "   "
		 << _(L"n:       ")
		 << n
		 << std::endl

                 << "   "
	//		 << _(L"χ²:         ")
                 << _(L"chi sq.: ")
                 << analyser.get_chi_square()
		 << std::endl

		 << "   "
		 << _(L"df:      ")
		 << analyser.get_df()
		 << std::endl

		 << "   "
		 << _(L"p:       ")
		 << analyser.get_p()
		 << std::endl
		 << std::endl
	
		 << std::endl
		 << "  "
		 << _(L"offsets:")
		 << std::endl ;
      
      const OffsetCounts & offset_counts(analyser.get_offset_counts());
      const OffsetProbabilities & offset_probabilities(analyser.get_offset_probabilities());
      OffsetCounts::const_iterator count;
      OffsetProbabilities::const_iterator prob;
      for (count = offset_counts.begin(), prob = offset_probabilities.begin();
	   count != offset_counts.end();
	   ++count, ++prob)
	{
	  std::wcout << L"   ["
		     << std::setiosflags(std::ios_base::fmtflags(0))
		     << std::setw(3)
		     << std::right
		     << count->first
		     << L"]:\t"

		     << L"p: "
		     << std::fixed
		     << std::setprecision(4)
		     << std::setw(6)
		     << prob->second
		     << L'\t'

		     << "count: "
		     << std::setiosflags(std::ios_base::fmtflags(0))
		     << std::setw(5)
		     << std::right
		     << count->second
		     << L'\t'


		     << L"ratio: "
		     << std::fixed
		     << std::setprecision(2)
		     << std::setw(5)
		     << std::right
		     << ((static_cast<double>(count->second)/static_cast<double>(n)) * 100.0)
		     << L'%'
		     << std::endl;
	}

      std::wcout << std::endl;

    } 
   catch (const NotEnoughResults & ner)
     {
       std::wcout << _(L"not enough results")
		  << std::endl;
       exit(0);
     }
   catch (const Db::Error & e)
     {
       std::wcerr << std::endl
		  << _(L"Database error: ");
       std::cerr  << e.what();
       std::wcerr << L") - aborting"
		  << std::endl;
       exit(1);
     }
   catch (const mysqlpp::BadQuery& e)
     {
       std::wcerr << std::endl
		  << _(L"Bad database query: ");
       std::cerr  << e.what()
		  << std::endl;
       exit(1);
     }
   catch (const mysqlpp::Exception& e)
     {
       std::wcerr << std::endl
		  << _(L"Unknown database error: ");
       std::cerr  << e.what()
		  << std::endl;
       exit(1);
     }

  exit (0);
}
