/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <time.h>

#include "libtkcommon/tkstring.h"
#include "libtkrandom/randomness.h"
#include "libtkrandom/retriever_registry.h"
#include "experimenter.h"
#include "text_frame.h"
#include "notes_entry.h"
#include "catcher.h"

using namespace TkResolver::Experimenter;

Experimenter:: Experimenter(const TkString & hostname,
			    const TkString & database,
			    const TkString & username,
			    const TkString & password)
  : _database(hostname, database, username, password),
    _parameters(&_database),
    _location(&_database, this),
    _operators(&_database, this),
    _notes(Gtk::TextBuffer::create()),
    //    _control(Gtk::Stock::MEDIA_RECORD, _toolbar.get_icon_size(), Gdk::Color("Blue"))
        _control(Gtk::Stock::GO_FORWARD, _toolbar.get_icon_size(), Gdk::Color("Blue"))
{
  set_title("TkResolver++ Experimenter");
  set_default_size(800, 600);
  set_position(Gtk::WIN_POS_CENTER);

  sigc::slot<void> execute_slot( sigc::mem_fun(*this, &Experimenter::execute_operators) );
  sigc::slot<void> control_slot( sigc::mem_fun(*this, &Experimenter::control_toggle) );
  sigc::slot<void> quit_slot   ( sigc::ptr_fun(&Gtk::Main::quit) );

  Gtk::VBox * box = Gtk::manage(new Gtk::VBox);
  add(*box);

  // toolbar
  box->pack_start(_toolbar, false, true);

  Gtk::ToolButton * tool_execute = Gtk::manage(new Gtk::ToolButton(Gtk::Stock::EXECUTE));
  _toolbar.append(*tool_execute, execute_slot);

  _toolbar.append(_control, control_slot);

  Gtk::ToolButton * tool_quit = Gtk::manage(new Gtk::ToolButton(Gtk::Stock::QUIT));
  _toolbar.append(*tool_quit, quit_slot);

  
  // toolbar tooltips
  Gtk::Tooltips * tooltips = _toolbar.get_tooltips_object();
  tooltips->set_tip(*tool_execute, "Execute single experiment");
  tooltips->set_tip(_control,      "Start control experiments");
  tooltips->set_tip(*tool_quit,    "Exit the experimenter");
  tooltips->enable();
  _toolbar.set_tooltips(true);
		  
  // parameters
  box->pack_start(_parameters, false, true);
  
  // location, operators and notes
  Gtk::VPaned * lists_and_notes = Gtk::manage(new Gtk::VPaned);
  box->pack_start(*lists_and_notes, true, true);
  
  Gtk::VPaned * lists = Gtk::manage(new Gtk::VPaned);
  lists_and_notes->pack1(*lists);
  lists->property_position() = 120;
  
  lists->pack1(_location,                 Gtk::FILL);
  lists->pack2(_operators, Gtk::EXPAND|Gtk::FILL);

  TextFrame * notes_frame = Gtk::manage(new TextFrame("Notes", _notes));
  lists_and_notes->pack2(*notes_frame, Gtk::FILL);
  

  // bottom execute button
  Gtk::Button * execute = Gtk::manage(new Gtk::Button(Gtk::Stock::EXECUTE));
  box->pack_start(*execute, false, true);
  execute->signal_clicked().connect(execute_slot);

  
  show_all_children();
  show();
}

namespace TkResolver
{
  class Execution : public Generator
  {
  private:
    const TkString &_retriever_type;
    const TkString &_algorithm_spec;
    const unsigned &_bit_descriptor;
    const unsigned short &_ntraces;
    const unsigned &_total_time;
    const double   &_speed;
    const double   &_horizon;
    unsigned *      _width;
    unsigned *      _height;
    time_t *        _start_date_time;
    Result *        _result;
    Random::Bits *  _used_bits;
  
  public:
    Execution (const TkString &retriever_type,
	       const TkString &algorithm_spec,
	       const unsigned &bit_descriptor,
	       const unsigned short &ntraces,
	       const unsigned &total_time,
	       unsigned &width,
	       unsigned &height,
	       const double &speed,
	       const double &horizon,
	       time_t &start_date_time,
	       Result &result,
	       Random::Bits &used_bits
	       )
      : _retriever_type(retriever_type),
	_algorithm_spec(algorithm_spec),
	_bit_descriptor(bit_descriptor),
	_ntraces(ntraces),
	_total_time(total_time),
	_width(&width),
	_height(&height),
	_speed(speed),
	_horizon(horizon),
	_start_date_time(&start_date_time),
	_result(&result),
	_used_bits(&used_bits) {}

    virtual void operator() ();
  };
} /* namespace TkResolver */

void
TkResolver::Execution:: operator() ()
{
  TKResolver * tkresolver =
    new TKResolver(_retriever_type,
		   _algorithm_spec,
		   _bit_descriptor,
		   _ntraces,
		   _total_time,
		   *_width,
		   *_height,
		   _speed,
		   _horizon);
  
  
  *_start_date_time = time(NULL);
  
  *_result = tkresolver->run();

  *_width     = tkresolver->get_display_width();
  *_height    = tkresolver->get_display_height();
  *_used_bits = tkresolver->get_used_bits();
  
  delete tkresolver;
}

void
Experimenter:: execute(const Db::IdList & operators,
		       const TkString * notes,
		       bool interactive)
{
  //
  // get parameters
  //
  TkString location_name;
  if (!get_location_name(location_name))
    return;

  unsigned width(_parameters.get_width());
  unsigned height(_parameters.get_height());
  TkString experiment_group(_parameters.get_experiment_group());

  TkString retriever_type;
  bool retriever_specified;
  _parameters.get_retriever_type(experiment_group, retriever_type, retriever_specified);
  

  // get experiment group's parameters from db 
  TkString algorithm_spec;
  unsigned bit_desc;
  unsigned short ntraces;
  unsigned total_time;
  double speed;
  double horizon;
  _database.experiment_group_get_parameters(experiment_group,
					    &algorithm_spec,
					    &bit_desc,
					    &ntraces,
					    &total_time,
					    &speed,
					    &horizon);

  // do it
  time_t start_date_time;
  Result result;
  Random::Bits used_bits;
  
  Execution executer(retriever_type,
		     algorithm_spec,
		     bit_desc,
		     ntraces,
		     total_time,
		     width,
		     height,
		     speed,
		     horizon,
		     start_date_time,
		     result,
		     used_bits);

  if ( catch_exception(executer, interactive) )
    return;
  
  // record result
  if (result.size() > 0) // user didn't escape
    {
      if (interactive && !notes)
	{
	  NotesEntry * notes_dialog = new NotesEntry(*this, _notes);
	  notes_dialog->run();
	  delete notes_dialog;
	}
      
      Random::RetrieverRegistry retrievers;
      _database.insert_experiment(experiment_group,
				  start_date_time,
				  location_name,
				  notes ? *notes : TkString(_notes->get_text()),
				  width,
				  height,
				  retriever_type,
				  retrievers.is_true_randomness(retriever_type),
				  retriever_specified,
				  used_bits,
				  operators,
				  result);
    }
}

bool
Experimenter:: get_location_name(TkString & location_name, bool interactive)
{
  try
    {
      location_name = _location.get_name();
    }
  catch (const Db::Ui::NoSelection & ns)
    {
      if (interactive)
	{
	  Gtk::MessageDialog dialog(*this, "No location selected", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
	  dialog.set_title("Execution error");
	  dialog.run();
	}
      return false;
    }

  return true;
}


void
Experimenter:: execute_operators()
{
  Db::IdList operators;
  try
    {
      operators = _operators.get_operators();
    }
  catch (const Db::Ui::NoSelection & ns)
    {
      Gtk::MessageDialog dialog(*this, "No operators selected", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
      dialog.set_title("Execution error");
      dialog.run();
      return;
    }

  execute(operators);
}

void
Experimenter:: control_toggle()
{
  if (_control.get_active())
    {
      if (!control_start())
	_control.set_active(false);
    }
  else
    control_stop();
}

bool
Experimenter:: control_start()
{
  TkString location_name; // we're not going to use this; just checking
  if (!get_location_name(location_name, false))
    return false;

  // anything between 10 minutes to 1 hour
  const unsigned lower_wait = 1000 * 60 * 10;
  const unsigned upper_wait = 1000 * 60 * 60;
  unsigned wait_length(lower_wait + Random::Randomness::get_number(upper_wait - lower_wait));
  WDEBUG_OUT << L"control waiting for " << wait_length << L" milliseconds (" << (wait_length/1000/60) << " minutes)" << std::endl;
  _timeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &Experimenter::control_notify), wait_length);

  return true;
}

bool
Experimenter:: control_notify()
{
  _control.flash_start(500);
  unsigned notify_time = 1000 * 30;
  _timeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &Experimenter::control_exec), notify_time);
  return false;
}

void
Experimenter:: control_stop()
{
  _timeout.disconnect();
  _control.flash_stop();
}

bool
Experimenter:: control_exec()
{
  DEBUG_OUT << "execing control" << std::endl;
  control_stop();
  
  Db::IdList null_operators;  

  TkString notes("Control");
  
  TkString user_notes(_notes->get_text());
  if (user_notes.size() > 0)
    notes += TkString(L" (") + user_notes + L")";

  execute(null_operators, &notes, false);

  control_start();
  return false;
}
