/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <sstream>
#include <boost/assign/list_of.hpp>

#include "operator_selector.h"

using namespace TkResolver;
using namespace Experimenter;
using namespace Db::Ui;

const static NameList selection_columns   = boost::assign::list_of(Name("id"));

OperatorSelector:: OperatorSelector(Db::Database * database,
					  Gtk::Window * parent)
  : DataSelectorList<OperatorInput> (database,
					"operator",
					selection_columns,
					parent,
					"Operator")
{
  const static NameList anonymous_columns   = boost::assign::list_of(Name("anonymous"));
  const static NameList operator_columns = boost::assign::list_of(Name("anonymous"))(Name("name"))(Name("email"));

  SlotListCellData * operator_column_slot =
    new SlotListCellData(sigc::mem_fun(*this, &OperatorSelector::render_operator_cell));

  const static ListColumnList display_columns = boost::assign::list_of
    (ListColumnInfo("Anonymous",   anonymous_columns,   0 ))
    (ListColumnInfo("Operator", operator_columns, operator_column_slot))
    ;

  set_display_columns(display_columns);

  get_view().get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
}

void
OperatorSelector:: render_operator_cell (Gtk::CellRenderer * renderer_super,
					       const Gtk::TreeModel::iterator iter)
{
  Gtk::CellRendererText * renderer = dynamic_cast<Gtk::CellRendererText *>(renderer_super);

  bool anonymous;
  iter->get_value(get_column_index("anonymous"), anonymous);

  if (anonymous)
    {
      Db::Id id;
      iter->get_value(get_column_index("id"), id);
      
      std::ostringstream strm;
      strm << "Operator #";
      strm << id;
      
      renderer->property_text() = Glib::ustring(strm.str());
    }
  else
    {
      Glib::ustring name, email;
      
      iter->get_value(get_column_index("name"), name);
      iter->get_value(get_column_index("email"), email);
      
      renderer->property_text() = name + " <" + email + ">";
    }
}

Db::IdList
OperatorSelector:: get_operators()
{
  Db::IdList id_list;
  Db::Ui::Selection selection(get_selection());
  for (Db::Ui::Selection::iterator iter(selection.begin());
       iter != selection.end();
       ++iter)
    {
      id_list.push_back((*iter)["id"]);
    }
  return id_list;
}


OperatorInput:: OperatorInput(Gtk::Window & parent)
  : DataInput("Operator details", parent),
    _male("Male"),
    _anonymous("Anonymous"),
    _send_results("Send results"),
    _notes(Gtk::TextBuffer::create())
{


  set_default_size(400, 400);

  Gtk::Table * table = Gtk::manage(new Gtk::Table(8, 2));
  get_vbox()->pack_start(*table, true, true);
  
  table->attach(*Gtk::manage(new Gtk::Label("Gender",               Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 0, 1, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Date of birth",        Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 1, 2, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Anonymity",            Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 2, 3, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Results availability", Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 3, 4, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Name",                 Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 4, 5, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Place of birth",       Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 5, 6, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Email",                Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 6, 7, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Notes",                Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 7, 8, Gtk::FILL, Gtk::FILL);
  

  Gtk::HBox * gender_box = Gtk::manage(new Gtk::HBox);
  table->attach(*gender_box, 1, 2, 0, 1, Gtk::FILL, Gtk::FILL);

  gender_box->pack_start(_male, false, true);
  Gtk::RadioButton::Group gender_group(_male.get_group());
  Gtk::RadioButton * female = Gtk::manage(new Gtk::RadioButton("Female"));
  female->set_group(gender_group);

  gender_box->pack_start(*female, false, true);

  
  // set the date back 25 years
  guint year, month, day;
  _date_of_birth.get_date(year, month, day);
  _date_of_birth.select_month(month, year - 25);

  _date_of_birth.display_options(Gtk::CALENDAR_SHOW_HEADING);
  table->attach(_date_of_birth,  1, 2, 1, 2, Gtk::FILL, Gtk::FILL);
  
  _anonymous.signal_toggled().connect( sigc::mem_fun(*this, &OperatorInput::update_anonymity) );
  table->attach(_anonymous,      1, 2, 2, 3, Gtk::FILL, Gtk::FILL);

  _send_results.set_active(true);
  table->attach(_send_results,   1, 2, 3, 4, Gtk::FILL, Gtk::FILL);

  table->attach(_name,           1, 2, 4, 5, Gtk::FILL, Gtk::FILL);
  table->attach(_place_of_birth, 1, 2, 5, 6, Gtk::FILL, Gtk::FILL);
  table->attach(_email,          1, 2, 6, 7, Gtk::FILL, Gtk::FILL);

  Gtk::ScrolledWindow * notes_scroll = Gtk::manage(new Gtk::ScrolledWindow);
  notes_scroll->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
  table->attach(*notes_scroll,   1, 2, 7, 8);
  
  Gtk::TextView * notes_view = Gtk::manage(new Gtk::TextView(_notes));
  notes_scroll->add(*notes_view);
  
  update_anonymity();
  show_all_children();
}


void
OperatorInput:: update_anonymity()
{
  std::list<Gtk::Widget *> identifying_widgets =
    boost::assign::list_of
    (static_cast<Gtk::Widget *>(&_send_results))
    (static_cast<Gtk::Widget *>(&_name))
    (static_cast<Gtk::Widget *>(&_place_of_birth))
    (static_cast<Gtk::Widget *>(&_email));

  bool sensitive = !_anonymous.get_active();
  for (std::list<Gtk::Widget *>::const_iterator iter(identifying_widgets.begin());
       iter != identifying_widgets.end();
       ++iter)
    (*iter)->set_sensitive(sensitive);
}

TkResolver::Db::Parameters *
OperatorInput:: get_data_sub()
{
  Db::Parameters * parameters = new Db::Parameters;

  parameters->insert( Db::Parameter("gender",         Db::String(_male.get_active() ? "M" : "F")) );
  
  guint year, month, day;
  _date_of_birth.get_date(year, month, day);
  Db::Date date(year, month + 1, day);
  std::ostringstream date_strm;
  date_strm << date;
  parameters->insert( Db::Parameter("date_of_birth",  Db::String(date_strm.str())) );
  
  parameters->insert( Db::Parameter("anonymous",      Db::String(_anonymous.get_active() ? 1 : 0)) );
  parameters->insert( Db::Parameter("send_results",   Db::String(_send_results.get_active() ? 1 : 0)) );
  parameters->insert( Db::Parameter("name",           Db::String(_name.get_text())) );
  parameters->insert( Db::Parameter("place_of_birth", Db::String(_place_of_birth.get_text())) );
  parameters->insert( Db::Parameter("email",          Db::String(_email.get_text())) );
  parameters->insert( Db::Parameter("notes",          Db::String(_notes->get_text())) );

  return parameters;
}
