/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2005 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <cmath>

#include "libtkcommon/debug.h"
#include "analyser.h"

extern "C" {
  double pochisq (double, int);
}

using namespace TkResolver;
using namespace Analysis;


class ResultAdaptor : public DataAdaptor
{
private:
  const Result &         _data;
  Result::const_iterator _iter;
  bool                   _done_first;
public:
  ResultAdaptor(const Result & data);
  virtual Offset get_offset();
  virtual bool emptied();
};

class DbAdaptor : public DataAdaptor
{
private:
  Db::DataCursor * _data;
  Db::Row          _row;
  bool             _done_first;
  bool             _emptied;

  void store_line();
public:
  DbAdaptor(Db::DataCursor * data);
  virtual Offset get_offset();
  virtual bool emptied();
};


Analyser:: Analyser(const OffsetProbabilities & offset_probabilities,
		    const Result & result_lines)
  : _offset_probabilities(offset_probabilities)
{
  ResultAdaptor adaptor(result_lines);
  analyse(adaptor);
}

Analyser:: Analyser(Db::Database & database,
		    const TkString & experiment_group,
		    bool pseudo,
		    bool control)
{
  TkString algorithm_spec;
  database.experiment_group_get_parameters(experiment_group, &algorithm_spec);

  Algorithm algorithm(algorithm_spec);
  _offset_probabilities = algorithm.get_probabilities();

  Db::DataCursor * data = database.experiment_group_get_results(experiment_group, pseudo, control);

  DbAdaptor adaptor(data);
  
  analyse(adaptor);

  delete data;
}

void
Analyser:: analyse (DataAdaptor & adaptor)
{
  if (adaptor.emptied())
    throw NotEnoughResults();

  // intialise the counts
  for (OffsetProbabilities::const_iterator iter = _offset_probabilities.begin();
       iter != _offset_probabilities.end();
       ++iter)
    {
      _offset_counts[iter->first] = 0;
    }

  // do the count
  _n = 0;
  Offset offset;
  do
    {
      offset = adaptor.get_offset();
      ++_offset_counts[offset];
      ++_n;
    }
  while (!adaptor.emptied());

  if (_n < 2)
    throw NotEnoughResults();

  //
  // calculate chi square
  //
  _chi_square = 0;
  Probability np;
  for (OffsetCounts::const_iterator iter = _offset_counts.begin();
       iter != _offset_counts.end();
       ++iter)
    {
      np = _n * _offset_probabilities[iter->first];
      _chi_square += ( powl(iter->second - np, 2) / np );
    }

  
  //
  // Use Perlman's numerical chi square stuff
  //
  _p = pochisq (_chi_square, get_df());
}

Probability
Analyser:: get_chi_square () const
{
  return _chi_square;
}

Probability
Analyser:: get_p () const
{
  return _p;
}

unsigned
Analyser:: get_df () const
{
  return _offset_probabilities.size() - 1;
}

Count
Analyser:: get_n () const
{
  return _n;
}

const OffsetCounts &
Analyser:: get_offset_counts() const
{
  return _offset_counts;
}

const OffsetProbabilities &
Analyser:: get_offset_probabilities() const
{
  return _offset_probabilities;
}




ResultAdaptor:: ResultAdaptor(const Result & data)
  : _data(data),
    _iter(_data.begin()),
    _done_first(false)
{
}

Offset
ResultAdaptor:: get_offset()
{
  if (!_done_first)
    {
      Offset o((*_iter)[0]);
      _done_first = true;
      return o;
    }
  else
    {
      Offset o((*_iter)[1]);
      ++_iter;
      _done_first = false;
      return o;
    }
}

bool
ResultAdaptor:: emptied()
{
  return !_done_first && (_iter == _data.end());
}

DbAdaptor:: DbAdaptor(Db::DataCursor * data)
  : _data(data),
    _done_first(false),
    _emptied(false)
{
  store_line();
}

void
DbAdaptor:: store_line()
{
  try
    {
      _row = _data->fetch_row();
    }
  catch (const mysqlpp::BadQuery & bq)
    {
      _emptied = true;
    }
  catch (const mysqlpp::EndOfResults & eor)
    {
      _emptied = true;
    }
}


Offset
DbAdaptor:: get_offset()
{
  if (!_done_first)
    {
      Offset o(_row.at(0));
      _done_first = true;
      return o;
    }
  else
    {
      Offset o(_row.at(1));
      store_line();
      _done_first = false;
      return o;
    }
}



bool
DbAdaptor:: emptied()
{
  return _emptied;
}
