/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <time.h>

#include <sstream>

#include "libtkcommon/mbs.h"
#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "database.h"

using namespace TkResolver;
using namespace Db;

static const char *
get_month_str(unsigned month)
{
  switch (month)
    {
    case  1: return "JAN";
    case  2: return "FEB";
    case  3: return "MAR";
    case  4: return "APR";
    case  5: return "MAY";
    case  6: return "JUN";
    case  7: return "JUL";
    case  8: return "AUG";
    case  9: return "SEP";
    case 10: return "OCT";
    case 11: return "NOV";
    case 12: return "DEC";
    }

  throw std::domain_error("unknown month number");
}

String
TkResolver::Db:: date_time_string(time_t t)
{
  struct tm *ltm;
  
  ltm = localtime (&t);
  if (!ltm)
    {
      throw std::runtime_error("could not convert time to local time components");
    }
  ltm->tm_year += 1900;
  ltm->tm_mon  += 1;

  std::ostringstream strm;
  strm << ltm->tm_year
       << "-"
       << (ltm->tm_mon < 10 ? "0" : "")
       << ltm->tm_mon
       << "-"
       << (ltm->tm_mday < 10 ? "0" : "")
       << ltm->tm_mday
       << " "
       << ltm->tm_hour
       << ":"
       << ltm->tm_min
       << ":"
       << ltm->tm_sec;

  return strm.str();
}


Error:: Error (mysqlpp::Connection & connection)
  : std::runtime_error(connection.error())
{
}


Database:: Database (const TkString & hostname,
		     const TkString & database,
		     const TkString & username,
		     const TkString & password)
  : _connection(std::string(database).c_str(),
		std::string(hostname).c_str(),
		std::string(username).c_str(),
		std::string(password).c_str())
{
  if (!_connection.connected())
    throw Error(_connection);

  create_table_info_query();
}

Database:: ~Database()
{
  delete _table_info_query;
}

void
Database:: create_table_info_query()
{
  _table_info_query = new mysqlpp::Query(&_connection);
  (*_table_info_query) << mysqlpp::SQLString("select * from %0:table");
  _table_info_query->parse();
}

void
Database:: get_table_info(const Name & name)
{
  mysqlpp::ResUse result = _table_info_query->use(mysqlpp::SQLString(name));
  if (!result)
    throw Error(_connection);

  const mysqlpp::FieldNames names (&result);
  const mysqlpp::FieldTypes types (&result);

  FieldInfo fields;
  unsigned  i = 0;
  for (mysqlpp::FieldNames::const_iterator name_iter = names.begin();
       name_iter != names.end();
       ++name_iter, ++i)
    {
      
      fields.insert(FieldInfo::value_type(*name_iter, types[i]));
    }

  _table_info.insert(TableInfo::value_type(name, fields));

  result.purge();
}

void
Database:: check_table_info(const Name & name)
{
  if (_table_info.find(name) == _table_info.end())
    get_table_info(name);
}

Type
Database:: get_column_type(const Name & table_name, const Name & column_name)
{
  check_table_info(table_name);

  const FieldInfo & fields(_table_info[table_name]);

  FieldInfo::const_iterator iter = fields.find(column_name);
  if (iter == fields.end())
    throw std::runtime_error(std::string("could not find column ") + table_name + "." + column_name);
  
  return iter->second;
}

DataSet *
Database:: get_table_data(const Name & table_name, const NameList & column_names)
{
  mysqlpp::Query query(&_connection);

  query << "select ";
  NameList::const_iterator iter, next;
  for (iter = column_names.begin();
       iter != column_names.end();
       iter = next)
    {
      next = iter;
      ++next;

      query << *iter;
      if (next != column_names.end())
	query << ",";
    }
  query << " from "
	<< table_name;

  DataSet * result = new DataSet(query.store());

  if (!(*result))
    throw Error(_connection);

  return result;
}

Id
Database:: insert_table_data(const Name & table_name, const Parameters & parameters)
{
  mysqlpp::Query query(&_connection);
  std::ostringstream values;
  mysqlpp::SQLQueryParms parms;

  query << "insert into " << table_name << "(";
  values << ") values (";

  Parameters::const_iterator iter, next;
  unsigned i = 0;
  for(iter = parameters.begin();
      iter != parameters.end();
      iter = next, ++i)
    {
      next = iter; ++next;

      query << iter->first;
      values << "%" << i << "r:" << iter->first;

      parms << iter->second;

      if(next != parameters.end())
	{
	  query << ",";
	  values << ",";
	}
    }
  query << values.str();
  query << ")";
  query.parse();

  mysqlpp::ResNSel result(query.execute(parms));
  if (!result)
    throw Error(_connection);

  return result.insert_id;
}

void
Database:: experiment_group_get_parameters(const TkString & group_name,
					   TkString * algorithm_spec,
					   unsigned * bit_desc,
					   unsigned short * traces,
					   unsigned * total_time,
					   double * speed,
					   double * horizon)
{
  mysqlpp::Query query(&_connection);

  query << " select algorithm_spec, bit_descriptor, traces, time, speed, horizon";
  query << " from experiment_group where name = %0r:group_name;";
  query.parse();

  query.def["group_name"] = std::string(group_name);

  DataSet result(query.store());

  if (!result)
    throw Error(_connection);


  unsigned field = 0;
  if (algorithm_spec)
    {
      std::string c_algorithm_spec( result.at(0).at(field++) );
      *algorithm_spec = c_algorithm_spec;
    }

  if (bit_desc)
    *bit_desc   = result.at(0).at(field++);
  if (traces)
    *traces     = result.at(0).at(field++);
  if (total_time)
    *total_time = result.at(0).at(field++);
  if (speed)
    *speed      = result.at(0).at(field++);
  if (horizon && speed && *speed > 0.0)
    *horizon    = result.at(0).at(field++);
}

Id
Database:: insert_experiment(const TkString & experiment_group,
			     time_t date_time,
			     const TkString & location_name,
			     const TkString & notes,
			     unsigned width,
			     unsigned height,
			     const TkString & retriever,
			     bool retriever_is_true_random,
			     bool retriever_specified,
			     const Random::Bits & randomness)
{
  Parameters params;

  params["group_name"]                  = experiment_group;
  params["date_time"]                   = date_time_string(date_time);
  params["location_name"]               = location_name;
  params["notes"]                       = notes;
  params["width"]                       = width;
  params["height"]                      = height;
  params["retriever"]                   = retriever;
  params["retriever_is_true_random"]    = retriever_is_true_random;
  params["retriever_specified"]         = retriever_specified;
  params["randomness"]                  = randomness.get_data();
  params["randomness_bits"]             = randomness.size();

  return insert_table_data("experiment", params);
}

void
Database:: insert_experiment(const Id & experiment_id,
			     const IdList & operators)
{
  for (IdList::const_iterator iter(operators.begin());
       iter != operators.end();
       ++iter)
    {
      Parameters params;
      
      params["experiment_id"]  = experiment_id;
      params["operator_id"] = *iter;
      
      insert_table_data("experiment_operator", params);
    }
}

void
Database:: insert_experiment(const Id & experiment_id,
			     const Result & result)
{
  unsigned ntraces( result.begin()->size() );
  unsigned set_index(0);
  unsigned offset_index;
  for (Result::const_iterator iter(result.begin());
       iter != result.end();
       ++iter, ++set_index)
    {
      Parameters result_params;
      result_params["experiment_id"] = experiment_id;
      result_params["set_index"]     = set_index;
      insert_table_data("experiment_result", result_params);


      for (offset_index = 0; offset_index < ntraces; ++offset_index)
	{
	  Parameters offset_params;
	  offset_params["experiment_id"] = experiment_id;
	  offset_params["set_index"]     = set_index;
	  offset_params["offset_index"]  = offset_index;
	  offset_params["offset"]        = (*iter)[offset_index];
	  insert_table_data("experiment_result_offset", offset_params);
	}
    }
}

Id
Database:: insert_experiment(const TkString & experiment_group,
			     time_t date_time,
			     const TkString & location_name,
			     const TkString & notes,
			     unsigned width,
			     unsigned height,
			     const TkString & retriever,
			     bool retriever_is_true_random,
			     bool retriever_specified,
			     const Random::Bits & randomness,
			     const IdList & operators,
			     const Result & result)
{
  Id id = insert_experiment(experiment_group,
			    date_time,
			    location_name,
			    notes,
			    width,
			    height,
			    retriever,
			    retriever_is_true_random,
			    retriever_specified,
			    randomness);

  insert_experiment(id, operators);
  insert_experiment(id, result);
  
  return id;
}


// choose a retriever from a valid set of retriever names, round-robin style
TkString
Database:: choose_retriever(const TkString & group_name, const TkStringList & retrievers)
{
  mysqlpp::Query query(&_connection);
  query << " select   retriever"
        << " ,        count(*) c"
	<< " from     experiment"
	<< " where    group_name = %0r:group_name"
	<< " group by retriever";
  query.parse();
  query.def["group_name"] = std::string(group_name);
  
  DataSet result(query.store());
  if (!result)
    throw Error(_connection);

  unsigned num_counts(result.num_rows());
  // if we have no counts, return the first retriever
  if (num_counts < 1)
    return retrievers.front();


  // search for retrievers that haven't been used
  for (TkStringList::const_iterator retr_iter(retrievers.begin());
       retr_iter != retrievers.end();
       ++retr_iter)
    {
      bool found = false;
      for (unsigned i = 0; i < num_counts; ++i)
	{
	  std::string count_retr(result.at(i).at(0));
	  if (*retr_iter == string_to_wstring(count_retr))
	    {
	      found = true;
	      break;
	    }
	}

      if (!found)
	{
	  return *retr_iter;
	}
    }

  // find the minimum used
  TkString min_retr ( string_to_wstring(std::string(result.at(0).at(0))) );
  unsigned     min_count( result.at(0).at(1) );

  for (unsigned i = 1; i < num_counts; ++i)
    {
      unsigned count(result.at(i).at(1));
      if (count < min_count)
	{
	  // check it exists in the list given
	  TkString retr( string_to_wstring(std::string(result.at(i).at(0))) );
	  for (TkStringList::const_iterator retr_iter(retrievers.begin());
	       retr_iter != retrievers.end();
	       ++retr_iter)
	    {
	      if (*retr_iter == retr)
		{
		  min_retr  = retr;
		  min_count = count;
		  break;
		}
	    }


	}
    }

  return min_retr;
}

DataCursor *
Database:: experiment_group_get_results(const TkString & group_name,
					bool true_random,
					bool control)
{
  mysqlpp::Query query(&_connection);
  query << " select   ("
	<< "             select   offset"
	<< "             from     experiment_result_offset"
	<< "             where    experiment_id  = e.id"
	<< "             and      set_index      = er.set_index"
	<< "             and      offset_index   = 0"
	<< "          ) offset_1"
	<< " ,        ("
	<< "             select   offset"
	<< "             from     experiment_result_offset"
	<< "             where    experiment_id  = e.id"
	<< "             and      set_index      = er.set_index"
	<< "             and      offset_index   = 1"
	<< "          ) offset_2"
	<< " from     experiment e"
	<< " ,        experiment_result er"
	<< " where    e.group_name               = %0r:group_name"
        << " and      e.retriever_is_true_random = %1r:true_random"
	<< " and      er.experiment_id           = e.id"
	<< " and      ("
	<< "             select   count(*)"
	<< "             from     experiment_operator"
	<< "             where    experiment_id  = e.id"
	<< "          )";
  if (control)
    query << "          = 0";
  else
    query << "          > 0";
 
  query.parse();

  query.def["group_name"]  = std::string(group_name);
  query.def["true_random"] = true_random;

  DataCursor * result = new DataCursor(query.use());
  if (!(*result))
    throw Error(_connection);

  return result;
}
