/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <sstream>

#include "data_selector.h"

using namespace TkResolver;
using namespace Db::Ui;


NameList
TkResolver::Db::Ui:: name_list_merge(const NameList & one, const NameList & two)
{
  NameList merged(one);
  bool exists;
  
  for (NameList::const_iterator iter_val(two.begin());
       iter_val != two.end();
       ++iter_val)
    {
      exists = false;
      for (NameList::const_iterator iter_check(merged.begin());
	   iter_check != merged.end();
	   ++iter_check)
	{
	  if (*iter_val == *iter_check)
	    exists = true;
	}

      if (!exists)
	merged.push_back(*iter_val);
    }

  return merged;
}

NameList
TkResolver::Db::Ui:: name_list_uniq(const NameList & names)
{
  NameList null_list;
  return name_list_merge (null_list, names);
}


NoSelection:: NoSelection()
  : std::runtime_error("No selection")
{
}



DataSelector:: DataSelector(Db::Database * database,
			    const Name & table_name,
			    const NameList & selection_columns)
  : _database(database),
    _table_name(table_name),
    _selection_columns(selection_columns)
{
}



void 
DataSelector:: set_columns(const NameList & display_columns)
{
  NameList columns(name_list_merge(_selection_columns, display_columns));
  
  Gtk::TreeModelColumnRecord column_record;
  Gtk::TreeModelColumnBase * column;
  
  for (NameList::const_iterator iter(columns.begin());
       iter != columns.end();
       ++iter)
    {
      column = create_model_column(*iter);
      _store_columns.push_back(TreeModelColumnList::value_type(*iter, column));
      column_record.add(*column);
    }

  _store = Gtk::ListStore::create(column_record);
  if(!_store)
    throw std::runtime_error("could not create store for data selector");

  refresh();
}

Name
DataSelector:: get_column_type_name(const Name & column_name) const
{
  Db::Type type(_database->get_column_type(_table_name, column_name));
  std::string type_name(type.base_type().name());

  DEBUG_OUT << "type for column "
	    << column_name
	    << ": "
	    << type_name
	    << " (from sql type '"
	    << type.sql_name()
	    << "')"
	    << std::endl;

  return Name(type_name);
}

Gtk::TreeModelColumnBase *
DataSelector:: create_model_column(const Name & name) const
{
  Name type_name(get_column_type_name(name));

  if (type_name == "Ss")
    return new Gtk::TreeModelColumn<Glib::ustring>;
  else if (type_name == "a")
    return new Gtk::TreeModelColumn<bool>;
  else if (type_name == "y")
    return new Gtk::TreeModelColumn<Db::Id>;
  else
    throw std::runtime_error(std::string("cannot create Gtk::TreeModelColumn from type '")
			     + type_name + "' for column " + name.c_str());
}


Gtk::CellRenderer *
DataSelector:: create_cell(const Name & column_name) const
{
  Name type_name(get_column_type_name(column_name));

  if (type_name == "Ss" || type_name == "y")
    {
      return Gtk::manage(new Gtk::CellRendererText);
    }
  else if (type_name == "a")
    {
      return Gtk::manage(new Gtk::CellRendererToggle);
    }
  else
    throw std::runtime_error(std::string("cannot create cell renderer from type '")
			     + type_name + "' for column " + column_name);
}

Name
DataSelector:: get_column_view_attribute(const Name & column_name) const
{
  Name type_name (get_column_type_name(column_name));

  if (type_name == "Ss" || type_name == "y")
    {
      return "text";
    }
  else if (type_name == "a")
    {
      return "active";
    }

  throw std::runtime_error(std::string("cannot create cell renderer from type '")
			   + type_name + "' for column " + column_name);
}





TreeModelColumnList::const_iterator
DataSelector:: get_store_column(const Name & name) const
{
  for (TreeModelColumnList::const_iterator iter(_store_columns.begin());
       iter != _store_columns.end();
       ++iter)
    {
      if(name == iter->first)
	{
	  return iter;
	}
    }

  throw std::logic_error(std::string("unknown column ") + name.c_str());
}


Gtk::TreeModelColumnBase *
DataSelector:: get_model_column(const Name & name) const
{
  return get_store_column(name)->second;
}


unsigned
DataSelector:: get_column_index(const Name & name) const
{
  return std::distance(_store_columns.begin(), get_store_column(name));
}




static void
set_column_value(Gtk::TreeModel::iterator & iter, int column_index, const Db::Value & value)
{
  Db::Type type(value.type());
  std::string type_name(type.base_type().name());

  if (type_name == "Ss")
    {
      Glib::ustring column_value = std::string(value);
      iter->set_value(column_index, column_value);
      return;
    }
  else if (type_name == "y")
    {
      Db::Id column_value = value;
      iter->set_value(column_index, column_value);
      return;
    }
  else if (type_name == "a")
    {
      bool column_value = value;
      iter->set_value(column_index, column_value);
      return;
    }

  throw std::domain_error(std::string("unknown column type '") + type_name + "' in instantiated DataSelector");
}



void
DataSelector:: append_row(const Db::Row & row)
{
  Gtk::TreeModel::iterator iter(_store->append());
  unsigned col_count = row.size();
  unsigned i;

  for (i = 0; i < col_count; ++i)
    set_column_value(iter, i, row.at(i));
}

Db::NameList
DataSelector:: store_db_columns()
{
  Db::NameList list;

  for (TreeModelColumnList::const_iterator iter(_store_columns.begin());
       iter != _store_columns.end();
       ++iter)
    {
      list.push_back(iter->first);
    }

  return list;

}


void
DataSelector:: refresh()
{
  Db::DataSet * data = _database->get_table_data(_table_name, store_db_columns());

  _store->clear();

  unsigned row_count = data->size();
  for (unsigned i = 0; i < row_count; ++i)
    {
      append_row(data->at(i));
    }

  delete data;

  refreshed();
  //  _view.get_selection()->select(Gtk::TreePath("0"));
}

Db::Value
DataSelector:: get_column_value(Gtk::TreeModel::iterator & iter, const Glib::ustring & column_name) const
{
  Name type_name (get_column_type_name(column_name));
  int column_index = get_column_index(column_name);

#define return_val(type) \
type val; \
iter->get_value(column_index, val); \
std::ostringstream strm; \
strm << val; \
return Db::Value(strm.str().c_str());

  if (type_name == "Ss")
    {
      /*      Glib::ustring val;
      iter->get_value(column_index, val);
      std::ostringstream strm;
      strm << val;
      return Db::Value(strm.str().c_str());*/
      return_val(Glib::ustring);
    }
  else if (type_name == "y")
    {
      return_val(Db::Id);
    }
  else if (type_name == "a")
    {
      return_val(bool);
    }

  throw std::domain_error(std::string("unknown column type '") + type_name + "' in instantiated DataSelector");
}

Selection
DataSelector:: get_selection() const
{
  Selection sel;
  ModelIteratorList selections(get_selected());

  for (ModelIteratorList::iterator sel_iter(selections.begin());
       sel_iter != selections.end();
       ++sel_iter)
    {
      Db::Values values;
      for (NameList::const_iterator col_iter(_selection_columns.begin());
	   col_iter != _selection_columns.end();
	   ++col_iter)
	{
	  values.insert(Db::Values::value_type(Db::Name(*col_iter), get_column_value(*sel_iter, *col_iter)));
	}

      sel.push_back(values);
    }
  
  return sel;
}

#define accessor(type, func, property) \
type \
DataSelector:: func() const \
{ \
  return property; \
}

accessor(const NameList &,                     get_selection_columns, _selection_columns);
accessor(const Glib::RefPtr<Gtk::ListStore> &, get_store,             _store);
accessor(const Name &,                         get_table_name,        _table_name);
accessor(Db::Database *,                       get_database,          _database);
