/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <stdexcept>
#include <cmath>

#include "libtkcommon/debug.h"
#include "libtkcommon/maths.h"
#include "decision_tree.h"
#include "randomness.h"

using namespace TkResolver;
using namespace Random;

DecisionTree:: DecisionTree(unsigned noutcomes, unsigned bit_descriptor)
  : _outcomes( noutcomes ),
    _bit_desc( bit_descriptor ),
    _depth( find_next_power_of_two_power(_outcomes) ),
    _retries(0)
{
  if (bit_descriptor % 2 == 0)
    throw std::invalid_argument("bit descriptor must be odd");
  
  // recursively generate tree
  unsigned outcome = 0;
  _root = new Node (_outcomes, _depth, outcome);
}

DecisionTree:: ~DecisionTree ()
{
  delete _root;
}

unsigned
DecisionTree:: get_depth() const
{
  return _depth;
}

unsigned
DecisionTree:: get_retries() const
{
  return _retries;
}

unsigned
DecisionTree:: tree_outcomes() const
{
  return 1U << _depth;
}

double
DecisionTree:: expected_tries() const
{
  unsigned outcomes(tree_outcomes());
  
  double expected_tries = 1.0;
  if (outcomes != _outcomes)
    {
      // approximating an infinite function
      const unsigned iterations = 15;
      double retry_rate( (outcomes - _outcomes) / static_cast<double>(outcomes) );
      double retries(0.0);

      for (unsigned i = 1; i < iterations; ++i)
	retries += pow(retry_rate, i);

      expected_tries += retries;
    }

  return expected_tries;
}

bool
DecisionTree:: deterministic() const
{
  return tree_outcomes() == _outcomes;
}

unsigned
DecisionTree:: generate_event(Randomness &rand)
{
  return _root->get_outcome(_root, _bit_desc, rand, _retries);
}


DecisionTree::Node:: Node(unsigned outcomes, unsigned depth, unsigned &outcome)
  : _outcome (0)
{
  if (depth == 0)
    {
      _children[0] = _children[1] = 0;
      _outcome = new unsigned(outcome++);
    }
  else
    {
      _children[0] = new Node(outcomes, depth - 1, outcome);
      if (outcome < outcomes)
	_children[1] = new Node(outcomes, depth - 1, outcome);
      else
	_children[1] = 0;
    }
}

DecisionTree::Node:: ~Node ()
{
  if (_outcome)
    delete _outcome;
  if (_children[0])
    delete _children[0];
  if (_children[1])
    delete _children[1];
}

static bool
decide(unsigned bit_descriptor, Randomness& rand)
{
  unsigned count(0);

  for (unsigned i = 0; i < bit_descriptor; ++i)
    {
      count += rand.get_bit();
    }

  return count > ((bit_descriptor - 1) / 2);
}

unsigned
DecisionTree::Node:: get_outcome(Node * root, unsigned bit_descriptor, Randomness& rand, unsigned &retries) const
{
  if (_outcome)
    return *_outcome;

  if ( decide(bit_descriptor, rand) )
    {
      return _children[0]->get_outcome(root, bit_descriptor, rand, retries);
    }
  else
    {
      if (!_children[1])
	{
	  ++retries;
	  return root->get_outcome(root, bit_descriptor, rand, retries);
	}
      else
	return _children[1]->get_outcome(root, bit_descriptor, rand, retries);
    }
}

