/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <wchar.h>

#include <iostream>
#include <string>

#include <SDL/SDL.h>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/real_time.h"
#include "libtkrandom/random_types.h"
#include "tkresolver.h"
#include "types.h"

using namespace TkResolver;

TKResolver:: TKResolver(const TkString & retriever_type,
			const TkString & algo,
			unsigned bit_descriptor,
			unsigned short ntraces,
			unsigned total_time,
			unsigned width,
			unsigned height,
			double   speed,
			double   horizon)
  : _ntraces(ntraces),
    _total_time(total_time),
    _display(new Display (height, width, _ntraces, _total_time, speed, horizon)),
    _algorithm(algo),
    _decision_tree(_algorithm.get_outcome_count(), bit_descriptor),
    _rand(retriever_type,
	  _decision_tree.deterministic(),
	  ((_display->get_total_lines() * _ntraces * bit_descriptor) / Random::BITS_PER_CHAR) + 1,
	  true),
    _current_line( _display->get_start_line() )
{
}

TKResolver::~TKResolver()
{
  delete _display;
}

unsigned
TKResolver:: get_display_height() const
{
  return _display->get_height();
}

unsigned
TKResolver:: get_display_width() const
{
  return _display->get_width();
}

unsigned
TKResolver:: get_total_lines() const
{
  return _display->get_total_lines();
}

const Random::Bits&
TKResolver:: get_used_bits() const
{
  return _rand.get_used_bits();
}

bool
TKResolver:: control()
{
  SDL_Event event;

  while (SDL_PollEvent (&event))
    {
      switch (event.type)
	{
	case SDL_QUIT:
	  return true;

	case SDL_KEYUP:
	  switch (event.key.keysym.sym)
	    {
	    case SDLK_ESCAPE:
	      return true;

	    default:
	      break;
	    }
	  
	default:
	  break;
	}
      
    }

  return false;
}


void
TKResolver:: slice()
{
  Line line;
  int offset;
  for (unsigned short i = 0; i < _ntraces; ++i)
    {
      offset = _algorithm.choose_offset( _decision_tree.generate_event(_rand) );
      line.push_back(offset);
      _current_line[i] += offset;
    }
  
  _display->print_line(_current_line);
  
  _offsets.push_back(line);
}

static Result null_result;

const Result&
TKResolver:: run()
{
  thread_set_real_time(80);

  _display->count_down(5);

  tk_time_t slice_size;
  tk_time_t slice_start;
  long      wait_size;
  int       i;
  bool      user_quit;
  long      timing_error(0);
  
  slice_size = (_total_time * MILLISECONDS_PER_SECOND) / _display->get_total_lines();
  
  for (i = 0; i < _display->get_total_lines(); i++)
    {
      slice_start = SDL_GetTicks();
      
      user_quit = control();
      if (user_quit)
	return null_result;
      
      slice ();
      
      wait_size = slice_size - (SDL_GetTicks() - slice_start);
      if (wait_size > 0)
	{
	  SDL_Delay(wait_size);
	}
      else
	{
	  timing_error += -wait_size;
	  if (timing_error > 10000) // milliseconds
	    throw std::runtime_error("display is too slow");
	}
    }

  return _offsets;
}
